/*
 *  MPEG-1 Real Time Encoder
 *
 *  Copyright (C) 1999-2001 Michael H. Schimek
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* $Id: options.c,v 1.25 2005/06/29 21:21:53 mschimek Exp $ */

#include "site_def.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#ifdef HAVE_GETOPT_LONG
#include <getopt.h>
#endif
#include <ctype.h>
#include <limits.h>
#include <sys/time.h>
#include <asm/types.h>
#include <linux/soundcard.h>
#include "common/videodev2.h"
#include "common/types.h"
#include "common/math.h"
#include "common/log.h"
#include "common/fifo.h"
#include "video/video.h"
#include "audio/libaudio.h"
#include "audio/mpeg.h"
#include "options.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

static const char *mux_options[] = { "", "video", "audio", "video_and_audio", "subtitles" };
static const char *mux_syn_options[] = { "nirvana", "bypass", "mpeg1", "mpeg2-ps", "vcd" };
static const char *audio_options[] = { "stereo", "", "dual_channel", "mono" };
static const char *mute_options[] = { "unmute", "mute", "ignore" };
static const char *cpu_options[] = { "", "pmmx", "p2", "p3", "p4", "k6-2", "k7", "m2", "c3", "n1" };
static const char *skip_options[] = { "compatible", "mux", "fake_picture" };

void usage(FILE *fi);

void
usage(FILE *fi)
{
	fprintf(fi,
		"Real time MPEG-1 encoder " VERSION "\n"
		"Copyright (C) 1999-2002 Michael H. Schimek\n"
		"\n"
		"This is free software licensed without a fee under the terms of the\n"
		"GNU General Public License Version 2. NO WARRANTIES.\n"
		"\n"
		"Usage: %s [options]\n\n"
		"Option                                                      Default\n"
		" -m mode        Encode 1 = video, 2 = audio, 3 = both       %s\n"
		" -X mux         Multiplexer to use                          auto\n"
		"\n"
		" -b bps         Output video bits per second                %5.3f Mbits/s\n"
		" -c name        Video capture device (V4L2/V4L API)         %s\n"
		" -F mode        Filter mode                                 auto\n"
		" -f frames      Frames per second                           maximum\n"
		" -g string      Group of pictures sequence (display order)  %s\n"
		" -l             Letterbox mode                              off\n"
		" -w             Grey mode                                   off\n"
		" -n frames      Number of video (audio) frames to encode\n"
		"    h:m[:s]     or until termination (Ctrl-C). To break\n"
		"                immediately hit Ctrl-\\                      years\n"
		" -s wxh         Image size (centred)                        %d x %d pixels\n"
		" -G wxh         Grab size, multiple of 16 x 16              %d x %d pixels\n"
//		" -H frames      Repeat sequence header every n frames,\n"
//		"                n > 0. Helps random access                  2 secs\n"
		" -R min,max     Motion compensation search range limits     %d,%d\n"
#if TEST_PREVIEW && HAVE_LIBXV
		" -P             XvImage Preview (test mode)                 disabled\n"
#endif
		"\n"
		" -I name        VBI device for Teletext subtitles           %s\n"
		" -T page        Teletext subtitle page(s), e.g. 1?0,200     none\n"
		"\n"
		" -a mode        Audio mode 0 = stereo, 2 = dual channel,\n"
		"                3 = mono                                    %s\n"
		" -p name        PCM sampling device                         %s\n"
		" -B bps         Output audio bits per second                %d kbits/s\n"
		" -S rate        Audio sampling rate                         %2.1f kHz\n"
		"\n"
		" -r line,vol    Audio record source 1..30%s,\n"
		"                volume 0..100                               %d,%d\n"
		" -x name        Audio mixer device (OSS API)                %s\n"
		" -M mode        RF audio 0 = unmute, 1 = mute, 2 = ignore   %s\n"
		"\n"
		" -i filename    Source configuration file\n"
		" -v             Increase verbosity level, try -v, -vv\n"
		" -C type        CPU type (p2, p3, k6-2, k7, ...)            auto\n"
		"\n"
                " -o name        Output file name                            stdout\n"
		"\n"
		"A configuration file can be piped in from standard input. See the\n"
		"mp1e manual page for details.\n",

		program_invocation_name, mux_options[modules], (double) video_bit_rate / 1e6,
		cap_dev, gop_sequence, width, height, grab_width, grab_height,
		motion_min, motion_max,

		vbi_dev,

		audio_options[audio_mode], pcm_dev, audio_bit_rate / 1000, sampling_rate / 1e3,

		mix_sources(), mix_line, mix_volume, mix_dev, mute_options[mute]
		);

	exit((fi == stderr) ? EXIT_FAILURE : EXIT_SUCCESS);
}

#define OPT_STR "02a:b:c:d:e:f:g:hi:jk:lm:n:o:p:r:s:t:vwx:zA:C:B:F:G:H:I:J:K:L:M:PR:S:T:VX:"

static const struct option
long_options[] = {
	{ "mpeg2",			no_argument,	   NULL, '0' }, /* testing */
	{ "hack2",			required_argument, NULL, '2' },
	{ "audio_mode",			required_argument, NULL, 'a' },
	{ "video_bit_rate",		required_argument, NULL, 'b' },
	{ "capture_device",		required_argument, NULL, 'c' },
	{ "cap_buffers",		required_argument, NULL, 'd' },
	{ "skip_method",		required_argument, NULL, 'e' },
	{ "frame_rate",			required_argument, NULL, 'f' },
	{ "gop_sequence",		required_argument, NULL, 'g' },
	{ "help",			no_argument,	   NULL, 'h' },
	{ "config",			required_argument, NULL, 'i' },
	{ "cut_output",			no_argument,	   NULL, 'j' },
	{ "break",			required_argument, NULL, 'k' },
	{ "letterbox",			no_argument,	   NULL, 'l' },
	{ "mux_mode",			required_argument, NULL, 'm' },
	{ "modules",			required_argument, NULL, 'm' }, /* alias */
	{ "num_frames",			required_argument, NULL, 'n' },
	{ "outputfile",			required_argument, NULL, 'o' },
	{ "pcm_device",			required_argument, NULL, 'p' },
	{ "rec_source",			required_argument, NULL, 'r' },
	{ "image_size",			required_argument, NULL, 's' },
	{ "test",			required_argument, NULL, 't' },
	{ "verbose",			optional_argument, NULL, 'v' },
	{ "mono",			no_argument,	   NULL, 'w' },
	{ "mixer_device",		required_argument, NULL, 'x' },
	{ "split",			no_argument,	   NULL, 'z' },
	{ "anno",			required_argument, NULL, 'A' },
	{ "audio_bit_rate",		required_argument, NULL, 'B' },
	{ "cpu",			required_argument, NULL, 'C' },
	{ "filter",			required_argument, NULL, 'F' },
	{ "grab_size",			required_argument, NULL, 'G' },
	{ "frames_per_seq_header",	required_argument, NULL, 'H' },
	{ "vbi_device",			required_argument, NULL, 'I' },
        { "source_fps",                 required_argument, NULL, 'J' }, /* AIW */
        { "half_rate",                  required_argument, NULL, 'K' }, /* AIW */
	{ "sample_aspect",		required_argument, NULL, 'L' },
	{ "mute",			required_argument, NULL, 'M' },
	{ "preview",			required_argument, NULL, 'P' },
	{ "motion",			required_argument, NULL, 'R' },
	{ "sampling_rate",		required_argument, NULL, 'S' },
	{ "subtitle_pages",		required_argument, NULL, 'T' },
	{ "version",			no_argument,	   NULL, 'V' },
	{ "mux",			required_argument, NULL, 'X' },
	{ NULL }
};

static rte_bool parse_option(int c);

static void
options_from_file(char *name, rte_bool fail)
{
	static int recursion = 0;
	char *s, *p, buffer[300];
	FILE *fi;
	int i;

	if (recursion++ > 20)
		FAIL("Recursion in option file"); // probably

	if (!strcmp(name, "stdin"))
		fi = stdin;
	else if (!(fi = fopen(name, "r")) && !fail) {
		printv(3, "Cannot open '%s' (ignored)\n", name);
		return;
	}

	ASSERT("open options file '%s'", fi != NULL, name);

	printv(2, "Opened '%s'\n", name ? name : "stdin");

	while (fgets(buffer, sizeof(buffer) - 2, fi)) {
		for (s = buffer; *s && *s != '#'; s++);
		*s = 0;

		while (s > buffer && isspace(s[-1])) s--;
		*s = 0;

		for (s = buffer; isspace(*s); s++);

		if (!*s)
			continue;

		p = s;

		while (!isspace(*s)) s++;
		while (isspace(*s)) *s++ = 0;
		if (*s == '=') *s++ = 0;
		while (isspace(*s)) s++;

		optarg = *s ? s : NULL;

		for (i = 0; long_options[i].name; i++)
			if (!strcmp(p, long_options[i].name))
				break;

		if (!long_options[i].val)
			FAIL("Unrecognized option '%s' in file '%s'", p, name);

		if (!optarg && long_options[i].has_arg == required_argument)
			FAIL("Option '%s' in file '%s' requires an argument", p, name);

		if (!parse_option(long_options[i].val))
			FAIL("Incorrect parameters for option '%s' in file '%s'", p, name);
	}

	ASSERT("read error in file '%s'", !ferror(fi), name);

	if (strcmp(name, "stdin"))
		fclose(fi);

	recursion--;
}

static int
suboption(const char **optp, int n, int def)
{
	int i;

	if (optarg) {
		if (isdigit(optarg[0])) {
			i = strtol(optarg, NULL, 0);
			if (i < 0 || i >= n)
				FAIL("Invalid parameter\n");
			return i;
		}

		for (i = 0; i < n; i++) {
			if (!optp[i] || !optp[i][0])
				continue;

			if (!strcmp(optarg, optp[i]))
				return i;
		}

		return -1;
	}

	return def;
}

static int
multi_suboption(const char **optp, int n, int def)
{
	char *s = optarg;
	int r = 0, l, i;

	if (!s || !*s)
		return def;

	while (isspace(*s))
		s++;

	if (!*s)
		return -1;

	do {
		while (isspace(*s))
			s++;

		if (!*s)
			return r;

		if (isdigit(*s))
			r |= strtol(s, &s, 0);
		else {
			for (i = 0; i < n; i++) {
				if (!optp[i] || !optp[i][0])
					continue;

				l = strlen(optp[i]);

				if (strncmp(s, optp[i], l) == 0
				    && (s[l] == 0 || isspace(s[l])
					|| strchr("+&|,", s[l]))) {
					s += l;
					r |= i;
					break;
				}
			}

			if (i >= n)
				return -1;
		}

		while (isspace(*s))
			s++;

	} while (*s && strchr("+&|,", *s++));

	return r;
}

static rte_bool have_audio_bit_rate = FALSE;
static rte_bool have_image_size = FALSE;
static rte_bool have_grab_size = FALSE;
static rte_bool have_letterbox = FALSE;
static rte_bool have_filter = FALSE;

static rte_bool
parse_option(int c)
{
	int i;

	switch (c) {
	case '0':
#ifdef OPTIONS_M2I
		m2i = 1;
#endif
		break;

	case '2':
		skip_method = 1; /* compatibility */
		break;

	case 'a':
		if ((audio_mode = suboption(audio_options, 4, audio_mode)) < 0)
			return FALSE;
		/*
		 *  0 = stereo, 1 = joint stereo, 2 = dual channel, 3 = mono;
		 *  To override psychoacoustic analysis trade-off:
		 *  +0  level 0 (no analysis)
		 *  +10 level 1
		 *  +20 level 2 (full analysis)
		 *
		 *  Higher levels are forced at low bit rates.
		 */
		break;

	case 'b': // video bit rate (bits/s, kbit/s or Mbits/s)
	{
		double rate = strtod(optarg, NULL);
			
		if (rate >= 20e3)
			video_bit_rate = rate;
		else if (rate >= 20)
			video_bit_rate = rate * 1e3;
		else
			video_bit_rate = rate * 1e6;

		if (video_bit_rate < 32000 ||
		    video_bit_rate > 20e6)
			return FALSE;

		break;
	}

	case 'c':
		cap_dev = strdup(optarg);
		break;

	case 'd':
		cap_buffers = strtol(optarg, NULL, 0);
		break;

	case 'e':
		if ((skip_method = suboption(skip_options, 3, 0)) < 0)
			return FALSE;
		break;

	case 'f':
		frame_rate = strtod(optarg, NULL);
		if (frame_rate <= 0.0)
			return FALSE;
		break;

	case 'g':
	{
		int i, len = strlen(optarg);

		if (len == 0)
			break;

		gop_sequence = strdup(optarg);

		for (i = 0; i < len; i++)
			gop_sequence[i] = toupper(gop_sequence[i]);

		break;
	}

	case 'm':
		modules = multi_suboption(mux_options, 5, modules);
		if (modules <= 0 || modules > 7)
			return FALSE;
		break;

	case 'n':
		if (strchr(optarg, ':')) {
			char *s = optarg;

			video_num_secs = 3600 * strtol(s, &s, 0);
			while (*s && !isalnum(*s)) s++;
			if (*s) {
				video_num_secs += 60 * strtol(s, &s, 0);
				while (*s && !isalnum(*s)) s++;
				if (*s)
					video_num_secs += strtol(s, &s, 0);
			}
			audio_num_secs = video_num_secs;
		} else {
			audio_num_frames = strtod(optarg, NULL);
			video_num_frames = audio_num_frames; 
		}
		break;

	case 'h':
		usage(stdout);
		break;

	case 'i':
		options_from_file(optarg, TRUE);
		break;

	case 'j':
		cut_output = !cut_output;
		break;

	case 'k':
		part_length = 1024 * (long long) strtol(optarg, NULL, 0);
		if (part_length < 0)
			usage(stdout);
		break;

	case 'l':
		have_letterbox = TRUE;
		break;

	case 'o':
		strncpy(outFile, optarg, sizeof(outFile) - 1);
		break;

	case 'p':
		pcm_dev = strdup(optarg);
		break;

	case 'r':
		sscanf(optarg, "%d,%d", &mix_line, &mix_volume);
		if (mix_line < 0 || mix_line > 30 ||
		    mix_volume < 0 || mix_volume > 100)
			return FALSE;
		break;

	case 's':
		have_image_size = TRUE;
		sscanf(optarg, "%dx%d", &width, &height);
		if (width < 1 || width > MAX_WIDTH ||
		    height < 1 || height > MAX_HEIGHT)
			return FALSE;
		break;

	case 't':
		/*
		  Test routines (values can be added):
		  1 - audio fft, filter and mdct test
		  2 - video
		  4 - DONT USE
		  8 - video frame dropping test (20%)
		  16 - video effective bit rate
		  32 - deliberate sampling rate mismatch
		  64 - store raw audio data (driver interface test)
		  128 - randomize video
		  0x100 - enable clock drift code
		  0x200 - fixed video pattern, silent audio
		*/
		test_mode = strtol(optarg, NULL, 0);
		break;

	case 'v':
		if (optarg)
			verbose = strtol(optarg, NULL, 0);
		else
			verbose++;
		break;

	case 'w':
		luma_only = !luma_only;
		break;

	case 'x':
		mix_dev = strdup(optarg);
		break;

	case 'z':
		split_sequence = TRUE;
		break;

	case 'A':
		anno = strdup(optarg);
		break;

	case 'B': // audio bit rate (bits/s or kbits/s)
	{
		double rate = strtod(optarg, NULL);

		if (rate >= 1e3)
			audio_bit_rate = rate;
		else
			audio_bit_rate = rate * 1e3;

		have_audio_bit_rate = TRUE;

		break;
	}

	case 'C':
		if ((cpu_type = suboption(cpu_options, 10, 0)) < 0)
			return FALSE;
		break;

	case 'F':
		filter_mode = strtol(optarg, NULL, 0);
		if (filter_mode <= 0 ||
		    filter_mode >= CM_NUM_MODES)
			return FALSE;
		have_filter = TRUE;
		break;

	case 'G':
		have_grab_size = TRUE;
		sscanf(optarg, "%dx%d", &grab_width, &grab_height);
		if (width < 1 || width > MAX_WIDTH ||
		    height < 1 || height > MAX_HEIGHT)
			return FALSE;
		break;

	case 'H':
		break;

		/* 
		   obsolete, ignored
			   
		   frames_per_seqhdr = strtol(optarg, NULL, 0);
			   
		   if (frames_per_seqhdr < 1)
		   return FALSE;
		   .BI "\-H, \-\-frames_per_seq_header " n
		   .RS
		   The sequence header in an MPEG-1 video stream contains information
		   such as the picture size and rate, and is therefore essential for a
		   player. For archiving it's sufficient to encode only one header.
		   If you want to stream the file, the player may have missed the initial
		   header. This option allows to repeat the header at certain regular
		   intervals, the default is appx. two seconds. When you concatenate
		   MPEG streams (simply with "cat") using different compression
		   parameters, you may also want to repeat the header to ensure random
		   access will pick up the correct parameters.
		   .P
		   Note whatever you specify the header is not inserted more often than
		   once in each group of pictures (see above). A reasonable value may
		   be 25 or 30, that is one header per second.
		   .RE
		   .TP
		*/

	case 'I':
		vbi_dev = strdup(optarg);
		break;

	case 'J':
		source_fps = strtod(optarg, NULL);
		if (source_fps <= 0.0)
			return FALSE;
		break;

	case 'K':
		if (optarg) {
			fix_interlaced = strtol(optarg, NULL, 0);
			if ((fix_interlaced < 1)||(fix_interlaced > 3))
				return FALSE;
			}
		else
			fix_interlaced = 1;
                break;

	case 'L':
		sample_aspect = strtod (optarg, NULL);
		break;

	case 'M':
		mute = suboption(mute_options, 3, mute);
		if (mute < 0 || mute > 2)
			return FALSE;
		break;

	case 'P':
#if !TEST_PREVIEW
		FAIL("Not configured for preview\n");
#endif
		preview++;
		break;

	case 'R':
		motion_max = 0;
		sscanf(optarg, "%d,%d", &motion_min, &motion_max);
		if (!motion_max)
			motion_max = motion_min;
		if (!motion_min)
			break;
		if (!have_filter)
			filter_mode = CM_YUYV_VERTICAL_INTERPOLATION;
		for (i = 0; i < 2; i++) {
			if (motion_min > motion_max)
				swap(motion_min, motion_max);
			if (motion_min < 4)
				motion_min = 4;
			if (motion_max > 64)
				motion_max = 64;
		}
		break;

	case 'S': // audio sampling rate (Hz or kHz)
	{
		double rate = strtod(optarg, NULL);

		if (rate >= 1e3)
			sampling_rate = rate;
		else
			sampling_rate = rate * 1e3;

		break;
	}

	case 'T':
		subtitle_pages = strdup(optarg);
		break;

	case 'V':
		puts("mp1e" " " VERSION " (" __DATE__ ")");
		exit(EXIT_SUCCESS);

	case 'X':
		mux_syn = suboption(mux_syn_options, 5, 2);
		if (mux_syn < 0 || mux_syn > 4)
			return FALSE;
		break;

	default:
		usage(stderr);
	}

	return TRUE;
}

static void
bark(void)
{
	if ((audio_mode & 3) != AUDIO_MODE_MONO && !have_audio_bit_rate)
		audio_bit_rate = audio_bit_rate_stereo;

	if (!have_image_size && have_grab_size) {
		width = grab_width;
		height = grab_height;
	} else if (have_image_size && !have_grab_size) {
		grab_width = width;
		grab_height = height;
	} else if (width > grab_width || height > grab_height)
		FAIL("Image size must be <= grab size\n");

	if (have_letterbox)
		height = height * 3 / 4; /* 16:9 */

	if (gop_sequence[0] != 'I' ||
	    strspn(gop_sequence, "IPB") != strlen(gop_sequence) ||
	    strlen(gop_sequence) > 1023) /* XXX 0P can increase the size */
		FAIL("Invalid group of pictures sequence: \"%s\".\n"
		     "A valid sequence can consist of the picture types 'I' (intra coded),\n"
		     "'P' (forward predicted), and 'B' (bidirectionally predicted) in any\n"
		     "order headed by an 'I' picture.", gop_sequence);

	if (audio_num_frames > 1000 && audio_num_frames < DBL_MAX && verbose > 0)
		fprintf(stderr, "Hint: you can write -n h:m[:s] instead of -n %lld\n",
			(long long) audio_num_frames);

	have_audio_bit_rate = FALSE;
	have_image_size = FALSE;
	have_grab_size = FALSE;
	have_letterbox = FALSE;
	have_filter = FALSE;
}

void
options(int ac, char **av)
{
	int index, c;
	char *s, *mp1erc = NULL;

	if ((s = getenv("HOME"))) {
		if ((mp1erc = malloc(strlen(s) + 16))) {
			strcpy(mp1erc, s);
			strcat(mp1erc, "/.mp1erc");
		}
	}

	options_from_file("/usr/local/etc/mp1e.conf", FALSE);

	if (mp1erc) {
		options_from_file(mp1erc, FALSE);
		free(mp1erc);
	}

	bark();

	if (!isatty(STDIN_FILENO)) {
		/* HvdH: ugly hack; check to see if the capture device is 'raw:'
		   in that case, stdin if for raw YUV data, not options.
		   cleaner solution is to reverse order; command line first,
		   then stdin options. It makes sense to have stdin override
		   command line options, but it might break stuff, so I won't 
		   change the order */ 
		for (c = 0; c < ac; c++) {
			/* FIXME: very rough check for capture device */
			if (strstr(av[c], "raw:")) {
				c = -1;
				break;
			}
		}

		if (c >= 0)
			options_from_file("stdin", FALSE);
	}

	while ((c = getopt_long(ac, av, OPT_STR, long_options, &index)) != -1)
		if (!parse_option(c))
			usage(stderr);

	bark();

	if (width * height < 128000)
		cap_buffers *= 2;

//	if (motion_min && motion_max)
//		cap_buffers *= 2;

	if (split_sequence && part_length)
		FAIL("-z|--split and -k|--break do not combine.");

	if ((split_sequence || part_length) && !outFile[0])
		FAIL("You must specify an output file name with the\n"
		     "-o option when using the -z|--split or -k|--break option.");
}
